/*
 * herdstat -- src/action_find_handler.cc
 * $Id: action_find_handler.cc 801 2005-11-02 15:00:00Z ka0ttic $
 * Copyright (c) 2005 Aaron Walker <ka0ttic@gentoo.org>
 *
 * This file is part of herdstat.
 *
 * herdstat is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * herdstat is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * herdstat; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 325, Boston, MA  02111-1257  USA
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <map>
#include <utility>
#include <algorithm>
#include <iterator>

#include <herdstat/portage/find.hh>
#include <herdstat/portage/exceptions.hh>

#include "pkgcache.hh"
#include "action_meta_handler.hh"
#include "action_find_handler.hh"

using namespace herdstat;
using namespace herdstat::portage;
using namespace herdstat::util;

action_find_handler::~action_find_handler()
{
}

int
action_find_handler::operator() (opts_type &opts)
{
    if (options::all())
    {
        std::cerr << "find handler does not support the 'all' target."
            << std::endl;
        return EXIT_FAILURE;
    }

    opts_type results;
    pkgcache pkgcache(options::portdir());

    if (options::regex())
    {
        regexp.assign(opts.front());

        matches = portage::find_package_regex(regexp, options::overlay(),
                    &search_timer, pkgcache);

        if (matches.empty())
        {
            std::cerr << "Failed to find any packages matching '"
                << regexp() << "'." << std::endl;
            return EXIT_FAILURE;
        }
    }
    else
    {
        for (opts_type::iterator i = opts.begin() ; i != opts.end() ; ++i)
            matches.insert(std::make_pair("", *i));
    }

    std::multimap<std::string, std::string>::iterator m;
    for (m = matches.begin() ; m != matches.end() ; ++m)
    {
        std::pair<std::string, std::string> p;

        try
        {
            if (options::regex())
                p = *m;
            else
                p = portage::find_package(m->second, options::overlay(),
                        &search_timer, pkgcache);
        }
        catch (const portage::AmbiguousPkg &e)
        {
            /* ambiguous still matches */
            results.insert(results.end(), e.packages.begin(), e.packages.end());
            continue;
        }
        catch (const portage::NonExistentPkg &e)
        {
            std::cerr << m->second << " doesn't seem to exist." << std::endl;

            if (matches.size() == 1)
                return EXIT_FAILURE;
            
            continue;
        }
        
        results.push_back(p.second);
    }

    if (results.size() > 1)
    {
        std::sort(results.begin(), results.end());
        results.erase(std::unique(results.begin(), results.end()),
            results.end());
    }

    if (options::meta())
    {
        /* disable stuff we've already handled */
        options::set_regex(false);
        options::set_eregex(false);
        
        action_meta_handler mhandler;
        mhandler(results);
    }
    else if (not options::count())
    {
        if (results.size() == 1)
            *stream << results.front() << std::endl;
        else
        {
            std::copy(results.begin(), results.end(), 
                std::ostream_iterator<std::string>(*stream, "\n"));
        }
    }

    size = results.size();
    flush();
    return EXIT_SUCCESS;
}

/* vim: set tw=80 sw=4 et : */
